from __future__ import annotations
from dataclasses import dataclass
import wisepy2
import pathlib
import base64
import ast_compat as ast
import contextlib
import io


@dataclass
class Context:
    registered_sources: dict[str, tuple[bool, str, bytes]]

    def register_module(
        self,
        module_path: list[str],
        src_path: pathlib.Path,
        source_code: str,
        *,
        is_package: bool,
        rootpath: pathlib.Path,
    ):
        if not module_path:
            raise ValueError("a Python package is required!")

        src_path = src_path.relative_to(rootpath)

        self.registered_sources[".".join(module_path)] = (
            is_package,
            str(src_path),
            base64.b64encode(source_code.encode("utf-8")),
        )


@contextlib.contextmanager
def _enter_path(xs: list[str], x: str):
    try:
        xs.append(x)
        yield xs
    finally:
        xs.pop()


def _traverse(
    rootpath: pathlib.Path,
    current_module_path: list[str],
    current_path: pathlib.Path,
    data: Context,
):
    if current_path.is_file() and current_path.suffix == ".py":
        if current_path.name == "__init__.py":
            return
        source = current_path.read_text(encoding="utf-8")
        name = current_path.name[: -len(".py")]
        if not name.isidentifier():
            return
        with _enter_path(current_module_path, name):
            data.register_module(
                current_module_path,
                current_path,
                source,
                is_package=False,
                rootpath=rootpath,
            )
    elif current_path.is_dir() and current_path.name.isidentifier():
        if current_path.name == "__pycache__":
            return
        entry = current_path / "__init__.py"
        name = current_path.name
        with _enter_path(current_module_path, name):
            if entry.exists():
                source = entry.read_text(encoding="utf-8")
                data.register_module(
                    current_module_path,
                    entry,
                    source,
                    is_package=True,
                    rootpath=rootpath,
                )
            else:
                data.register_module(
                    current_module_path, entry, "", is_package=True, rootpath=rootpath
                )
            for each in current_path.iterdir():
                _traverse(rootpath, current_module_path, each, data)


@wisepy2.wise
def CLI(*packages_or_modules: str, out: str, dynlinkloader: bool = False):
    ctx = Context({})
    for package_or_module in packages_or_modules:
        rootpath = pathlib.Path(package_or_module).absolute()
        if rootpath.is_file():
            rootpath = rootpath.parent
        p = pathlib.Path(package_or_module).absolute()
        _traverse(rootpath, [], p, ctx)
    buf = io.StringIO()
    if dynlinkloader:
        buf.write("from pypkgpack.importer import register_code_resource\n")
    else:
        buf.write(
            (pathlib.Path(__file__).parent / "importer.py").read_text(encoding="utf-8")
        )
    buf.write("\n")
    for fullname, (is_package, filepath, source_code) in ctx.registered_sources.items():
        node = ast.Call(
            ast.Name("register_code_resource", ctx=ast.Load()),
            [
                ast.Constant(fullname),
                ast.Constant(is_package),
                ast.Constant(filepath),
                ast.Constant(source_code),
            ],
        )
        ast.fix_missing_locations(node)
        buf.write(ast.unparse(node))
        buf.write("\n")
    with pathlib.Path(out).open("w", encoding="utf-8") as f:
        f.write(buf.getvalue())



from __future__ import annotations
from importlib.abc import MetaPathFinder, Loader
from importlib.machinery import ModuleSpec
import pathlib
import types
import base64
import warnings
import marshal
import hashlib
import sys
import os

__version__ = b"0.4.0"

BYTECODE_CACHE_DIR: pathlib.Path | None = pathlib.Path(
    os.getenv("PYBUNDLE_CACHE_PATH") or "~/.cache/pypkgpack"
).expanduser()

try:
    BYTECODE_CACHE_DIR.mkdir(mode=0o777, parents=True, exist_ok=True)
except IOError:
    # in case no permission
    BYTECODE_CACHE_DIR = None

FILE_PATH = str
IS_PACKAGE = bool
bundled_resources: dict[str, tuple[IS_PACKAGE, FILE_PATH, list[bytes]]] = {}
PYTHON_VERSION = ".".join(map(str, sys.version_info[:3]))


def register_code_resource(name: str, is_package: bool, filepath: str, *codes: bytes):
    if name in bundled_resources:
        old_is_package, old_filepath, old_codes = bundled_resources[name]
        old_codes.extend(codes)
        if filepath != old_filepath:
            warnings.warn(f"multiple '__file__' found for module {name}")
        if old_is_package is not is_package:
            warnings.warn(
                f"{name} has been asked to be {'package' if old_is_package else 'module'}"
            )
            bundled_resources[name] = is_package, old_filepath, old_codes
    else:
        bundled_resources[name] = (is_package, filepath, list(codes))


class BundledSourceLoader(Loader):
    def __init__(
        self, is_package: bool, fullname: str, filepath: str, source_codes: list[bytes]
    ):
        self.fullname = fullname
        self.source_codes = source_codes
        self.is_package = is_package
        self.filepath = filepath

    def resolve_filepath(self):
        return str(pathlib.Path(__file__).parent / self.filepath)

    def create_module(self, spec: ModuleSpec) -> types.ModuleType | None:
        module = types.ModuleType(self.fullname)
        module.__name__ = self.fullname
        module.__loader__ = self
        module.__file__ = self.resolve_filepath()
        if self.is_package:
            module.__package__ = module.__name__
            module.__path__ = []
        else:
            module.__package__ = ".".join(module.__name__.split(".")[:-1])
        return module

    def compile_module(self, source_codes: list[bytes]):
        code_objects: list[types.CodeType] = []

        def do_compile():
            for src in source_codes:
                src = base64.b64decode(src).decode("utf-8")
                code_objects.append(compile(src, self.filepath, "exec"))

        if BYTECODE_CACHE_DIR is None:
            # NO cache dir available
            do_compile()
        else:
            md5 = 1192634
            md5 += int.from_bytes(__version__, "little", signed=False)
            for source_code in source_codes:
                digest = hashlib.md5(source_code).hexdigest().encode()
                md5 ^= int.from_bytes(digest, "little", signed=False)
            hexdigest = hex(md5).encode()
            cache_file = BYTECODE_CACHE_DIR / (self.fullname + "@" + PYTHON_VERSION)
            if cache_file.exists() and cache_file.is_file():
                with cache_file.open("rb") as f:
                    cache_contents = f.read()
            else:
                cache_contents = b""
            if cache_contents.startswith(hexdigest):
                code_objects.extend(marshal.loads(cache_contents[len(hexdigest) :]))
            else:
                do_compile()
                try:
                    cache_file.parent.mkdir(mode=0o777, parents=True, exist_ok=True)
                    with cache_file.open("wb") as f:
                        f.write(hexdigest)
                        f.write(marshal.dumps(code_objects))
                except IOError:
                    pass
        return code_objects

    def exec_module(self, module: types.ModuleType):
        code_objects = self.compile_module(self.source_codes)
        for code_object in code_objects:
            exec(code_object, module.__dict__)


class BundledSourceFinder(MetaPathFinder):
    def find_spec(self, fullname: str, path, target=None):
        if fullname in bundled_resources:
            is_package, filepath, codes = bundled_resources[fullname]
            return ModuleSpec(
                fullname,
                BundledSourceLoader(
                    is_package, fullname=fullname, filepath=filepath, source_codes=codes
                ),
            )


sys.meta_path.insert(0, BundledSourceFinder())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pypkgpack']

package_data = \
{'': ['*']}

install_requires = \
['ast-compat>=0.11.1,<0.12.0', 'wisepy2>=1.3,<2.0']

entry_points = \
{'console_scripts': ['pypkgpack = pypkgpack.bundle:CLI']}

setup_kwargs = {
    'name': 'pypkgpack',
    'version': '0.4.0',
    'description': 'Bundling your Python project into a single Python file.',
    'long_description': "## PyPkgPack\n\n[![CI](https://github.com/Suzhou-Tongyuan/pypkgpack/actions/workflows/ci.yml/badge.svg)](https://github.com/Suzhou-Tongyuan/pypkgpack/actions/workflows/ci.yml)\n[![versions](https://img.shields.io/pypi/pyversions/pypkgpack.svg)](https://pypi.org/project/pypkgpack/#history)\n[![pypi](https://img.shields.io/pypi/v/pypkgpack.svg)](https://pypi.org/project/pypkgpack/)\n[![License](https://img.shields.io/badge/License-BSD_2--Clause-green.svg)](https://github.com/Suzhou-Tongyuan/pypkgpack/blob/main/LICENSE)\n\nBundling multiple Python packages into a single Python file.\n\nUsage:\n\n```shell\n> pypkgpack /path/to/mypackage1 /path/to/mypackage2 --out bundled_package.py\n\n> python\n\npython> import bundled_package\npython> from mypackage.mymodule import myfunction\npython> print(myfunction())\n```\n\nFeatures:\n\n- [x] bundling a Python package into a single Python file\n- [x] caching bytecode compilation and cache invalidation\n- [x] fixing the missing `__init__.py`\n- [x] allow multiple source code implementations for the same module\n- [x] respect Python's import semantics\n- [x] bundled modules need no extra dependencies\n- [ ] support bundling binaries and assets\n- [ ] allow lazy imports\n",
    'author': 'thautwarm',
    'author_email': 'twshere@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
